from ..configs.parameters import get_default_player_symbol, get_default_check_obvious_plays
from ..env.win_conditions import look_for_obvious_steps


class InterfaceAI:
    def __init__(self, symbole=None, check_obvious_plays=None):
        """Créer un joueur du symbole indiqué"""
        if symbole is not None:
            self.player = symbole
        else:
            self.player = get_default_player_symbol()

        if check_obvious_plays is not None:
            self.check_obvious_plays = check_obvious_plays
        else:
            self.check_obvious_plays = get_default_check_obvious_plays()

    @classmethod
    def get_default_params(cls):
        params = dict()
        params['player'] = get_default_player_symbol()
        params['check_obvious_plays'] = get_default_check_obvious_plays()

        return params

    def print(self):
        print()
        print('[AI] symbol = {}'.format(self.player))
        print('[AI] bias to obvious steps = {}'.format(self.check_obvious_plays))

    def play_witout_bias(self, grille):
        raise NotImplementedError

    def simulate_end_game(self, grille):
        raise NotImplementedError

    @classmethod
    def load_model(cls):
        print('Learner does not know how to load a trained model.')

    @classmethod
    def save_model(cls):
        print('Learner does not know how to save a trained model.')

    def play(self, grille):
        if self.check_obvious_plays:
            my_play, is_forced_play = self.play_with_bias(grille)
        else:
            my_play = self.play_witout_bias(grille)
            is_forced_play = None  # We don't know since we don't check whether it would be a forced play.
        return my_play, is_forced_play

    def set_params(self, dico):
        for elem in dico:
            setattr(self, elem, dico[elem])
        return

    def set_all_parameters_to_default(self):
        self.set_params(self.get_default_params())
        return

    def play_with_bias(self, grille):
        """Jouer de façon biaisée : vérifier s'il est possible de gagner en un coup avant toute réflexion"""
        mon_coup_urgent = look_for_obvious_steps(grille,
                                                 player_symbol=self.player,
                                                 opponent_symbol=self.get_opponent_symbol())

        is_forced_play = bool(mon_coup_urgent is not None)

        if is_forced_play:
            my_play = mon_coup_urgent
        else:
            my_play = self.play_witout_bias(grille)

        return my_play, is_forced_play

    def get_player_symbol(self):
        return self.player

    def get_opponent_symbol(self):
        return self.get_other_symbol(self.player)

    @staticmethod
    def get_other_symbol(symbole):
        """Passer du symbole d'un joueur au symbole de son adversaire"""
        if symbole == 'X':
            return 'O'
        else:
            return 'X'

    # noinspection PyPep8Naming
    def equalize_computing_resources(self, UCT_ai_instance):
        # Give the same computing resources to player X (UCT) and player O (InterfaceAI):
        try:
            self.check_obvious_plays = UCT_ai_instance.check_obvious_plays
        except AttributeError:
            print(
                'Equalization of computing resources failed: missing attributes from {}'.format(repr(UCT_ai_instance)))



class Node(object):
    """
    Tree structure

    Reference: http://stackoverflow.com/questions/3753665/python-tree-structure-and-numerical-codes/3753796#3753796
    """

    byname = {}

    def __init__(self, name, parent=None):
        self.name = name
        self.parent = parent
        self.children = []
        self.byname[name] = self
        if parent is None:  # root pseudo-node
            self.code = 0
        else:  # all normal nodes
            self.parent.children.append(self)
            self.code = len(self.parent.children)



# For AI interface


def get_default_player_symbol():
    return 'O'


def get_default_check_obvious_plays():
    """
    Parameter regarding obvious plays during games (not during end-game simulations)

    If you AI is not very good, it might help it. Otherwise, it could be better to set this setting to False and try to
    build a better AI with an increase number of Monte Carlo samples. Typically, in UCT vs. MC games, MC benefits from
    this setting, but UCT does not (because UCT's Tree Down algorithm with num_descentes_dans_arbre > num_columns(=7)
    leads to actually checking all the children of the current position, which includes all possible obvious win-steps.

    """
    check_obvious_plays = True
    return check_obvious_plays


# For random AI, which is used to simulate end games for every AI!


def get_default_bias_to_obvious_steps():
    """
    Parameter regarding obvious plays during end-game simulations (not during games)

    [Warning] Important advice: you might want to keep this boolean to False! Otherwise, there is a huge burden of
    computation, which hinders your ability to play with other parameters such as the number of Monte Carlo samples.

    It might sound worth it as the end-game evaluations should be more accurate. However, AI spends most of its
    computation power checking whether each *random* move might be a winning move during end-game simulations, which
    is not a smart way to "think" for an AI. The result is an AI which is less good at the game than it could be.

    AI plays "obvious steps" (one-step win, or one-step win-deny for the opponent), no matter this parameter.
    Only the end-game simulations rely on this parameter.
    """

    bias_to_obvious_steps = True
    return bias_to_obvious_steps


def get_default_max_num_steps_to_explore():
    """
    [Warning] Important advice: you might want to keep this value small. Otherwise, AI will spend a lot of moves to
    evaluate which player wins in each situation, and this will force you to decrease the number of Monte Carlo
    samples. The result is that the value of each board state might be less accurate due to a low num_tirages_MC.

    If max_num_steps_to_explore is set to None, then end-game simulations are simulated until the end (win/loss/draw).
    """

    max_num_steps_to_explore = 30
    return max_num_steps_to_explore


# For both Monte Carlo AI and UCT AI

# noinspection PyPep8Naming
def get_default_num_tirages_MC():
    """
    Number of Monte Carlo samples used to estimate the value of a board state.
    """

    # noinspection PyPep8Naming
    num_tirages_MC = 8
    return num_tirages_MC


def get_default_action_sampling_strategy():
    """
    If true, then we try each allowed first action at the root of the tree, in which case:
     - we spend as much computing resources to value each board state obtained after a first action,
     - the total number of MC samples may slightly differ from 'num_tirages_MC'.

    Otherwise, we randomly sample the first actions at the root of the tree, in which case:
     - we might over-sample or under-sample a few of them,
     - we have a better control of the total number of MC samples.
    """

    deterministic_sampling_of_actions_at_root = True
    return deterministic_sampling_of_actions_at_root


# Only for UCT AI

def get_default_num_descentes_dans_arbre():
    """
    UCT algorithm relies on Monte-Carlo Tree search. However, it differs from MCT because it works in a layered manner.

        repeat 'num_descentes_dans_arbre' times the following procedure:
            - Tree Down: a node is reached based on UCT search criteria,
            - Monte Carlo: the node board state is evaluated using num_tirages_MC samples,
            - Tree Up: the evaluation is propagated up in the search tree using UCT update rules for evaluations.

    """
    num_descentes_dans_arbre = 7
    return num_descentes_dans_arbre


def get_default_facteur_uct():
    """
    The following UCT parameter could be equal to 0. Make sure to use a very small value. Typically, between 0 and 0.3.
    """
    facteur_uct = 0.01
    return facteur_uct


# For main

def get_default_num_parties_jouees():
    """
    The number of AI self-play games to find out which of two AI is the best one.
    """
    num_parties_jouees = 10
    return num_parties_jouees


def main():
    import timeit

    duration = timeit.timeit("prepare_and_train('Random', 10)",
                             setup="from puissance4.training import prepare_and_train",
                             number=1)
    print('Time elapsed = {:.2f} s'.format(duration))

    return


if __name__ == "__main__":
    main()



import pathlib
from math import sqrt, log
from random import choice

from .mc import MC
from ..configs.parameters import get_default_num_descentes_dans_arbre, get_default_facteur_uct
from ..env.grille import Grille
from ..lib.node import Node


class UCT(MC):
    """Intelligence artificielle reposant sur l'algorithme Upper-Confidence-Tree"""

    def __init__(self, symbole='O', num_descentes_dans_arbre=None, facteur_uct=None):
        super().__init__(symbole)

        if num_descentes_dans_arbre is not None:
            self.num_descentes_dans_arbre = num_descentes_dans_arbre
        else:
            self.num_descentes_dans_arbre = get_default_num_descentes_dans_arbre()

        if facteur_uct is not None:
            self.facteur_uct = facteur_uct
        else:
            self.facteur_uct = get_default_facteur_uct()

        self.dict_num_visits_of_board_state = dict()
        self.dict_for_action_in_board_state = dict()
        self.tree = None

        self.data_path = 'data/'
        self.node_visit_filename = 'node_visit.txt'
        self.node_action_filename = 'node_action.txt'

    def load_model(self):
        # NB: It is apparently useless to load these variables for different games.

        try:

            import ast

            with open(self.data_path + self.node_visit_filename, 'r', encoding="utf8") as f:
                self.dict_num_visits_of_board_state = ast.literal_eval(f.readlines()[0])

            with open(self.data_path + self.node_action_filename, 'r', encoding="utf8") as f:
                self.dict_for_action_in_board_state = ast.literal_eval(f.readlines()[0])

        except FileNotFoundError:
            print('Model files cound not be loaded.')

        return

    def save_model(self):
        # NB: It is apparently useless to save these variables for different games.

        # Reference of the following line: https://stackoverflow.com/a/14364249
        pathlib.Path(self.data_path).mkdir(parents=True, exist_ok=True)

        with open(self.data_path + self.node_visit_filename, 'w', encoding="utf8") as f:
            print(self.dict_num_visits_of_board_state, file=f)

        with open(self.data_path + self.node_action_filename, 'w', encoding="utf8") as f:
            print(self.dict_for_action_in_board_state, file=f)

        return

    def get_default_params(self):
        params = super().get_default_params()
        params['num_descentes_dans_arbre'] = get_default_num_descentes_dans_arbre()
        params['facteur_uct'] = get_default_facteur_uct()

        return params

    def print(self):
        super().print()
        print("[Upper Confidence Tree] number of tree descents = {}".format(self.num_descentes_dans_arbre))
        print("[Upper Confidence Tree] UCT factor = {}".format(self.facteur_uct))
        return

    @classmethod
    def print_strategy(cls):
        return

    def play_witout_bias(self, grille):
        """Déterminer la meilleure action en fonction des résultats de l'exploration UCT"""
        # recevoir une position p
        initial_board_state = grille.get_name()
        # soit A l'arbre UCT vide à la racine près
        self.tree = Node(initial_board_state)
        # boucle : plusieurs descentes dans l'arbre
        for _ in range(self.num_descentes_dans_arbre):
            # soit N le noeud résultant d'une descente dans l'arbre (A, p)
            (N, current_player_at_N) = self.tree_down()
            # soit R l'évaluation de p
            grille_copiee = Grille()
            grille_copiee.copy_name(N.name)

            # noinspection PyPep8Naming
            R, _ = self.simuler_monte_carlo(grille_copiee, current_player_at_N)
            # Attention! R est la valeur pour 'current_player_at_N', pas forcément pour 'self.player'!

            # nous effectuons une remontée de l'arbre (A, N, R)
            self.tree_up(N, R, current_player_at_N)
        # renvoyer le coup fils (de la racine de A) qui a la meilleure valeur UCT
        meilleure_action = self.choisir_action_uct(grille)
        return meilleure_action

    def tree_down(self):
        """Descendre dans l'arbre UCT"""
        # Initialisation des variables (N, current_player, grille) avant la boucle while

        # soit N la racine de l'arbre
        # noinspection PyPep8Naming
        N = self.tree

        current_player = self.player

        grille = Grille()
        grille.copy_name(N.name)

        # boucle : fils non explorés de N
        while all([(N.name, i) in self.dict_for_action_in_board_state for i in grille.look_for_allowed_steps()]):
            # soit F le fils de N ayant la plus grande valeur UCT
            if len(grille.look_for_allowed_steps()) > 0:
                action = self.choisir_action_uct(grille)
                nouvel_etat = self.changer_etat_apres_transition(N.name, action, current_player)
                f = Node(nouvel_etat, N)
                f.code = action

                # noinspection PyPep8Naming
                N = f
                current_player = self.get_other_symbol(current_player)
                grille.copy_name(N.name)

            else:
                break

        # soir F un fils de N tiré au hasard parmi les fils non explorés
        mes_coups_possibles = grille.look_for_allowed_steps()
        if len(mes_coups_possibles) > 0:
            actions_inexplorees = [i for i in mes_coups_possibles if
                                   (N.name, i) not in self.dict_for_action_in_board_state]
            action = choice(actions_inexplorees)
            etat_inexplore = self.changer_etat_apres_transition(N.name, action, current_player)
            f = Node(etat_inexplore, N)
            f.code = action

            current_player = self.get_other_symbol(current_player)
        else:
            f = N
        return f, current_player

    # noinspection PyPep8Naming
    def tree_up(self, node_N, value_R, current_player_at_N):
        """Remonter dans l'arbre UCT"""
        current_player = current_player_at_N

        while node_N is not self.tree:
            parent_board_state = node_N.parent.name
            action_to_reach_N = node_N.code

            try:
                self.dict_num_visits_of_board_state[node_N.name] += 1
            except KeyError:
                self.dict_num_visits_of_board_state[node_N.name] = 1

            try:
                self.dict_for_action_in_board_state[(parent_board_state, action_to_reach_N)]['count'] += 1
            except KeyError:
                self.dict_for_action_in_board_state[(parent_board_state, action_to_reach_N)] = dict()
                self.dict_for_action_in_board_state[(parent_board_state, action_to_reach_N)]['count'] = 1

            symbol_for_parent_of_current_player = self.get_other_symbol(current_player)

            # R is the valuation of Node N, in the best interest of 'current_player_at_N'.
            # However, q is the valuation of the PARENT of 'current_player', which means that:
            # q = R if the PARENT of current player shares its symbol with Node N,
            # q = -R if the PARENT of current player does not share its symbol with Node N.
            #
            if current_player != current_player_at_N:  # <- La ligne qui m'a fait perdre beaucoup de temps.
                q = value_R
            else:
                q = - value_R

            try:
                # En effet, nous nous intéressons au parent de N, et non à N lui-même.
                mu_avant = self.dict_for_action_in_board_state[(parent_board_state, action_to_reach_N)]['score']
            except KeyError:
                mu_avant = 0

            n = self.dict_for_action_in_board_state[(parent_board_state, action_to_reach_N)]['count']
            mu = mu_avant + (1.0 / n) * (q - mu_avant)
            self.dict_for_action_in_board_state[(parent_board_state, action_to_reach_N)]['score'] = mu

            node_N = node_N.parent
            current_player = symbol_for_parent_of_current_player

        # Enfin, visite de la racine.
        try:
            self.dict_num_visits_of_board_state[node_N.name] += 1
        except KeyError:
            self.dict_num_visits_of_board_state[node_N.name] = 1

        return

    def choisir_action_uct(self, grille):
        """Choisir une action en utilisant le critère UCB"""
        etat = grille.get_name()
        meilleure_action = None
        meilleure_evaluation = None
        for action in grille.look_for_allowed_steps():
            if (etat, action) not in self.dict_for_action_in_board_state:
                continue
            recompense_moyenne = self.dict_for_action_in_board_state[(etat, action)]['score']
            num_etat_action = self.dict_for_action_in_board_state[(etat, action)]['count']
            num_etat = self.dict_num_visits_of_board_state[etat]
            evaluation = recompense_moyenne + self.facteur_uct * sqrt(1.0 * log(num_etat) / num_etat_action)
            if meilleure_evaluation is None or evaluation > meilleure_evaluation:
                meilleure_evaluation = evaluation
                meilleure_action = action
        if meilleure_action is None:
            print("Aucun coup admissible.")
        return meilleure_action

    @staticmethod
    def changer_etat_apres_transition(etat, action, joueur):
        """Déterminer l'état obtenu lorsque le joueur effectue l'action dans l'état donné"""
        grille_copiee = Grille()
        grille_copiee.copy_name(etat)
        grille_copiee.drop(joueur, action)
        nouvel_etat = grille_copiee.get_name()
        return nouvel_etat

    # noinspection PyPep8Naming
    def equalize_computing_resources(self, UCT_ai_instance):
        # Give the same computing resources to player X (UCT) and player O (UCT):
        super().equalize_computing_resources(UCT_ai_instance)
        try:

            # Override the value set by superclass MC
            self.num_tirages_MC = UCT_ai_instance.num_tirages_MC

            # Set values for class UCT
            self.num_descentes_dans_arbre = UCT_ai_instance.num_descentes_dans_arbre
            self.facteur_uct = UCT_ai_instance.facteur_uct

        except AttributeError:
            print(
                'Equalization of computing resources failed: missing attributes from {}'.format(repr(UCT_ai_instance)))



from .ai_interface import InterfaceAI
from ..configs.parameters import get_default_bias_to_obvious_steps, get_default_max_num_steps_to_explore


class AI(InterfaceAI):
    """Intelligence artificielle"""

    def __init__(self, symbole='O', bias_to_obvious_steps=None, max_num_steps_to_explore=None):
        super().__init__(symbole)

        if bias_to_obvious_steps is not None:
            self.bias_to_obvious_steps = bias_to_obvious_steps
        else:
            self.bias_to_obvious_steps = get_default_bias_to_obvious_steps()

        if max_num_steps_to_explore is not None:
            self.max_num_steps_to_explore = max_num_steps_to_explore
        else:
            self.max_num_steps_to_explore = get_default_max_num_steps_to_explore()

    def get_default_params(self):
        params = super().get_default_params()
        params['bias_to_obvious_steps'] = get_default_bias_to_obvious_steps()
        params['max_num_steps_to_explore'] = get_default_max_num_steps_to_explore()

        return params

    def print(self):
        super().print()
        print('[End-game simulations] bias to obvious steps = {}'.format(self.bias_to_obvious_steps))
        print('[End-game simulations] maximal number of steps = {}'.format(self.max_num_steps_to_explore))
        return

    def play_witout_bias(self, grille):
        """Jouer de façon non biaisée : renvoyer au hasard un coup possible"""
        return grille.get_random_allowed_step()

    def simulate_end_game(self, grille):
        """Simuler une fin de partie à partir de la grille initiale fournie"""

        player_who_last_played = self.get_opponent_symbol()

        step_counter = 0
        first_action = None

        while not (grille.check_victory()) and len(grille.look_for_allowed_steps()) > 0:
            current_player = self.get_other_symbol(player_who_last_played)

            if self.bias_to_obvious_steps:
                my_play, _ = self.play_with_bias(grille)
            else:
                my_play = self.play_witout_bias(grille)

            if step_counter == 0:
                first_action = my_play

            grille.drop(current_player, my_play)

            player_who_last_played = current_player

            step_counter += 1
            if self.max_num_steps_to_explore is not None and step_counter >= self.max_num_steps_to_explore:
                break

        if grille.check_victory():
            winner_symbol = player_who_last_played
        else:
            winner_symbol = 'draw'

        return winner_symbol, first_action

    # noinspection PyPep8Naming
    def equalize_computing_resources(self, UCT_ai_instance):
        # Give the same computing resources to player X (UCT) and player O (AI):
        super().equalize_computing_resources(UCT_ai_instance)
        try:
            self.bias_to_obvious_steps = UCT_ai_instance.bias_to_obvious_steps
            self.max_num_steps_to_explore = UCT_ai_instance.max_num_steps_to_explore
        except AttributeError:
            print(
                'Equalization of computing resources failed: missing attributes from {}'.format(repr(UCT_ai_instance)))



import operator

from .ai import AI
from ..configs.parameters import get_default_num_tirages_MC, get_default_action_sampling_strategy
from ..env.grille import Grille


class MC(AI):
    """Intelligence artificielle reposant sur des simulations Monte-Carlo"""

    # noinspection PyPep8Naming
    def __init__(self, symbole='O', num_tirages_MC=None, deterministic_sampling_of_actions_at_root=None):
        super().__init__(symbole)

        if num_tirages_MC is not None:
            self.num_tirages_MC = num_tirages_MC
        else:
            self.num_tirages_MC = get_default_num_tirages_MC()

        if deterministic_sampling_of_actions_at_root is not None:
            self.deterministic_sampling_of_actions_at_root = deterministic_sampling_of_actions_at_root
        else:
            self.deterministic_sampling_of_actions_at_root = get_default_action_sampling_strategy()

    def get_default_params(self):
        params = super().get_default_params()
        params['num_tirages_MC'] = get_default_num_tirages_MC()
        params['deterministic_sampling_of_actions_at_root'] = get_default_action_sampling_strategy()

        return params

    def print(self):
        super().print()
        print("[Monte Carlo] number of samples for grid valuation = {}".format(self.num_tirages_MC))
        self.print_strategy()
        return

    def print_strategy(self):
        print("[Monte Carlo] deterministic strategy for action sampling at the root = {}".format(
            self.deterministic_sampling_of_actions_at_root))
        return

    def play_witout_bias(self, grille):
        """Déterminer la meilleure action en fonction des résultats des simulations Monte-Carlo"""
        meilleure_action = None

        allowed_steps = grille.look_for_allowed_steps()

        if len(allowed_steps) == 1:
            meilleure_action = allowed_steps[0]
        elif len(allowed_steps) > 1:
            if self.deterministic_sampling_of_actions_at_root:
                meilleure_action = self.uniformly_sample_actions_at_root(grille, allowed_steps)
            else:
                meilleure_action = self.randomly_sample_actions_at_root(grille)

        return meilleure_action

    def randomly_sample_actions_at_root(self, grille, num_end_game_simulations=None):

        if num_end_game_simulations is None:
            num_end_game_simulations = self.num_tirages_MC

        # C'est au tour du joueur (A).
        current_player = self.get_player_symbol()

        _, meilleure_action = self.simuler_monte_carlo(grille, current_player, num_end_game_simulations)

        return meilleure_action

    def uniformly_sample_actions_at_root(self, grille, allowed_steps=None, num_end_game_simulations=None):

        if allowed_steps is None:
            allowed_steps = grille.look_for_allowed_steps()

        if num_end_game_simulations is None:
            num_end_game_simulations = 1 + (self.num_tirages_MC // len(allowed_steps))

        meilleure_action = None
        meilleure_evaluation = None

        for action in allowed_steps:

            grille_simulee = Grille(grille)
            # Le joueur (A) joue un coup.
            grille_simulee.drop(self.player, action)
            # C'est maintenant au tour de l'autre joueur (B).
            R, _ = self.simuler_monte_carlo(grille_simulee, self.get_opponent_symbol(), num_end_game_simulations)
            # Nous prenons l'opposé de la valeur simulée, car nous nous intéressons au joueur A.
            evaluation = - R

            if meilleure_evaluation is None or evaluation > meilleure_evaluation:
                meilleure_evaluation = evaluation
                meilleure_action = action

        return meilleure_action

    def simuler_monte_carlo(self, grille, current_player, num_end_game_simulations=None):
        """
        Evaluer une grille par des simulations Monte-Carlo de la fin de la partie

        Attention! La valeur renvoyée correspond à l'estimation pour "current_player", et pas nécessairement
        pour "self.player" !
        """

        if num_end_game_simulations is None:
            num_end_game_simulations = self.num_tirages_MC

        victory_stats_for_first_actions = self.core_process_monte_carlo(grille, current_player,
                                                                        num_end_game_simulations)

        best_grid_valuation, best_first_action = self.post_process_monte_carlo(victory_stats_for_first_actions,
                                                                               current_player)

        return best_grid_valuation, best_first_action

    def core_process_monte_carlo(self, grille, current_player, num_end_game_simulations=None):

        if num_end_game_simulations is None:
            num_end_game_simulations = self.num_tirages_MC

        victory_stats_for_first_actions = dict()

        # Warning: this is a random AI, so you might want to use values different from the defaults used by MC for:
        # - self.bias_to_obvious_steps
        # - self.max_num_steps_to_explore
        ai = AI(current_player, self.bias_to_obvious_steps, self.max_num_steps_to_explore)

        for _ in range(num_end_game_simulations):
            grille_simulee = Grille(grille)
            winner_symbol, first_action = ai.simulate_end_game(grille_simulee)

            try:
                victory_stats_for_first_actions[first_action][winner_symbol] += 1
            except KeyError:
                victory_stats_for_first_actions[first_action] = {'O': 0, 'X': 0, 'draw': 0}
                victory_stats_for_first_actions[first_action][winner_symbol] += 1

        return victory_stats_for_first_actions

    def post_process_monte_carlo(self, victory_stats_for_first_actions, current_player):

        grid_valuations_for_first_actions = self.find_grid_valuations_for_first_actions(victory_stats_for_first_actions,
                                                                                        current_player)

        best_first_action = self.find_best_first_action(grid_valuations_for_first_actions)

        best_grid_valuation = self.find_best_grid_valuation(grid_valuations_for_first_actions, best_first_action)

        return best_grid_valuation, best_first_action

    def find_grid_valuations_for_first_actions(self, victory_stats_for_first_actions, player_symbol):
        # Compute a score based on the stats of victories

        opponent_symbol = self.get_other_symbol(player_symbol)

        grid_valuations_for_first_actions = dict()
        for (action, num_victoires) in victory_stats_for_first_actions.items():

            try:
                action_score = (num_victoires[player_symbol] - num_victoires[opponent_symbol]) \
                               / (num_victoires[player_symbol] + num_victoires[opponent_symbol]
                                  + num_victoires['draw'])
            except ZeroDivisionError:
                action_score = 0

            grid_valuations_for_first_actions[action] = action_score

        return grid_valuations_for_first_actions

    @classmethod
    def find_best_first_action(cls, grid_valuations_for_first_actions):
        # Reference: https://stackoverflow.com/a/268285
        best_first_action = max(grid_valuations_for_first_actions.items(), key=operator.itemgetter(1))[0]
        return best_first_action

    def find_best_grid_valuation(self, grid_valuations_for_first_actions, best_first_action=None):
        if best_first_action is None:
            best_first_action = self.find_best_first_action(grid_valuations_for_first_actions)
        best_grid_valuation = grid_valuations_for_first_actions[best_first_action]
        return best_grid_valuation

    # noinspection PyPep8Naming
    def equalize_computing_resources(self, UCT_ai_instance):
        # Give the same computing resources to player X (UCT) and player O (MC):
        super().equalize_computing_resources(UCT_ai_instance)
        try:
            # By increasing 'num_tirages_MC', the TOTAL number of MC samples should be ABOUT the same as for UCT.
            # NB: - ABOUT the same for the deterministic case uniformly_sample_actions_at_root().
            #     - EXACTLY the same for the non-deterministic case randomly_sample_actions_at_root().
            self.num_tirages_MC = UCT_ai_instance.num_tirages_MC * UCT_ai_instance.num_descentes_dans_arbre

            # NB: grid valuation depends on another parameter ('num_end_game_simulations') which is:
            # - computed on the fly for the deterministic case, so that the TOTAL number of samples is about as expected
            # - equal to 'num_tirages_MC' for the non-deterministic case.

        except AttributeError:
            print(
                'Equalization of computing resources failed: missing attributes from {}'.format(repr(UCT_ai_instance)))



from random import shuffle

from .agent.uct import UCT
from .env.grille import Grille
from .lib.utils import get_possible_player_inputs, convert_player_input, convert_to_column_display


def main():
    play_now()

    return True


def play_now(load_previously_trained_model=False):
    # AI player

    ai_player = UCT()

    # Load

    if load_previously_trained_model:
        # Load learnt model
        ai_player.load_model()

    # Play

    play_versus_ai(ai_player)

    return


def play_versus_ai(ai_player):
    # X always starts.
    player_symbols = ['X', 'O']

    # Randomly shuffle symbols assigned to each player
    shuffle(player_symbols)

    # Decide which player is X and which is O.
    ai_player.player = player_symbols[0]
    human_player = player_symbols[1]

    ai_player.print()
    print()

    print('Human player is {}'.format(human_player))
    print()

    grille = Grille()

    current_symbol = None
    current_player_name = None

    step_counter = 0
    while not (grille.check_victory()) and len(grille.look_for_allowed_steps()) > 0:

        if step_counter % 2 == 0:
            current_symbol = 'X'
        else:
            current_symbol = 'O'

        if current_symbol == ai_player.player:
            current_player_name = 'artificial intelligence'

            if step_counter > 0:
                grille.show_grid()

            my_play, is_forced_play = ai_player.play(grille)

            grille.drop(current_symbol, my_play)

        else:
            current_player_name = 'human being'

            grille.show_grid()

            while True:
                user_input = input("Letter a through g  = drop disc, q = quit. \nYour move:")
                if user_input in get_possible_player_inputs(grille.width):
                    my_play = convert_player_input(user_input)
                    if grille.drop(current_symbol, my_play):
                        break

            is_forced_play = None

        step_counter += 1

        if is_forced_play is not None and is_forced_play:
            print('Obvious play by AI in ({}, {}) below.'.format(current_symbol,
                                                                 convert_to_column_display(my_play).upper()))

    if grille.check_victory():
        print('Winner = {} ({}) after {} steps'.format(current_symbol, current_player_name, step_counter))
    else:
        print('Draw after {} steps'.format(step_counter))

    grille.show_grid()

    return


if __name__ == "__main__":
    main()



def get_conversion_offset():
    return 97


def convert_to_column_display(col):
    return chr(col + get_conversion_offset())


def get_possible_player_inputs(width):
    return [convert_to_column_display(i) for i in range(width)]


def convert_player_input(player_input):
    return ord(player_input) - get_conversion_offset()



from random import choice


class Grille:
    """Représentation du plateau de jeu"""

    def __init__(self, grille_initiale=None):
        """Créer une grille vide, ou copier une grille existante"""
        self.width = 7
        self.height = 6
        self.empty_space = '.'
        self.top_row_no = 0
        self.sep_ligne = '\n'
        self.sep_colonne = ';'
        self.grid = [[self.empty_space] * self.width for _ in range(self.height)]

        if grille_initiale is not None:
            self.copy_name(grille_initiale.get_name())

    def get_name(self):
        """Renvoyer un texte représentant la grille de façon unique"""
        return self.sep_ligne.join([self.sep_colonne.join(liste) for liste in self.grid])

    def get_grid(self):
        return self.grid

    def set_grid(self, grid):
        self.grid = grid

    def wipe(self):
        self.__init__()

    def copy_name(self, name):
        """Créer la grille représentée par le texte"""
        listes = name.split(self.sep_ligne)
        self.grid = [l.split(self.sep_colonne) for l in listes]

    def copy(self, grille):
        # TODO
        self.set_grid(grille.get_grid())

    def show_grid(self, reverse_displayed_row_no=False):
        """Afficher la grille"""
        sep = ' '
        print()
        for row_no, row in enumerate(self.grid):
            displayed_row_no = row_no
            if reverse_displayed_row_no:
                displayed_row_no = self.height - row_no
            print(str(displayed_row_no) + '|' + sep.join(row))
        print('  ' + sep.join('-' for _ in range(self.width)))
        print('  ' + sep.join(chr(i + 65) for i in range(self.width)))
        print()
        return

    def drop(self, disc, col):
        """Placer un jeton dans la colonne"""
        ce_coup_est_possible = False
        for row in reversed(self.grid):
            if row[col] == self.empty_space:
                row[col] = disc
                ce_coup_est_possible = True
                break
        return ce_coup_est_possible

    def check_victory(self):
        """Vérifier si quatre jetons consécutifs du même joueur sont alignés"""
        for y in range(len(self.grid)):
            for x in range(len(self.grid[y])):
                if self.grid[y][x] != self.empty_space:
                    # Alignement indépendant de y
                    if self.is_far_from_right(x):
                        # Alignement horizontal, le noeud (x,y) étant le plus à gauche
                        if all(self.grid[y][x] == self.grid[y][x + i + 1] for i in range(3)):
                            return True
                    # Alignements qui dépendent de y
                    if self.is_far_from_bottom(y):
                        # Alignement vertical, le noeud (x,y) étant le plus haut
                        if all(self.grid[y][x] == self.grid[y + i + 1][x] for i in range(3)):
                            return True
                        # Alignement diagonal, le noeud (x,y) étant le plus haut et à gauche
                        if self.is_far_from_right(x):
                            if all(self.grid[y][x] == self.grid[y + i + 1][x + i + 1] for i in range(3)):
                                return True
                        # Alignement diagonal, le noeud (x,y) étant le plus haut et à droite
                        if self.is_far_from_left(x):
                            if all(self.grid[y][x] == self.grid[y + i + 1][x - i - 1] for i in range(3)):
                                return True
        return False

    def look_for_allowed_steps(self):
        """Renvoyer la liste des coups autorisés"""
        return [x for x in range(self.width) if self.grid[self.top_row_no][x] == self.empty_space]

    def get_random_allowed_step(self):
        """Renvoyer un coups au hasard parmi ceux autorisés"""
        return choice(self.look_for_allowed_steps())

    def get_num_steps(self):
        return sum([self.grid[y][x] != self.empty_space for y in range(self.height) for x in range(self.width)])

    def is_empty_space(self, x, y):
        return bool(self.grid[y][x] == self.empty_space)

    def is_above_occupied_space(self, x, y):
        return self.is_at_bottom(y) or not (self.is_empty_space(x, y + 1))

    def is_playable(self, x, y):
        return self.is_empty_space(x, y) and self.is_above_occupied_space(x, y)

    def play_if_possible(self, x, y):
        my_play = None
        if self.is_playable(x, y):
            my_play = x
        return my_play

    def is_at_bottom(self, y):
        return bool(y == self.height - 1)

    @staticmethod
    def is_at_top(y):
        return bool(y == 0)

    def is_very_far_from_bottom(self, y):
        return self.is_far_from_bottom(y, cushion=-1)

    def is_far_from_bottom(self, y, cushion=0):
        return bool(y < self.height - 3 + cushion)

    def is_quite_far_from_bottom(self, y):
        return self.is_far_from_bottom(y, cushion=1)

    @staticmethod
    def is_far_from_top(y):
        return bool(y > 2)

    def is_far_from_right(self, x):
        return bool(x < self.width - 3)

    @staticmethod
    def is_far_from_left(x):
        return bool(x > 2)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='puissance4',
    version='0.6.1',
    author='Wok',
    author_email='wok@tuta.io',
    description='Artificial Intelligence for the game Connect Four on PyPI',
    keywords=['puissance4', 'puissance-4', 'connect4', 'connect-4', 'connect-four', 'artificial intelligence', 'UCT'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/woctezuma/puissance4',
    download_url='https://github.com/woctezuma/puissance4/archive/0.6.1.tar.gz',
    packages=setuptools.find_packages(),
    install_requires=[
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Games/Entertainment',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: French',
    ],
    python_requires='>=3',
)



import unittest

import puissance4


class TestTrainingMethods(unittest.TestCase):

    def test_prepare_and_train(self):
        is_consistent, _, _ = puissance4.prepare_and_train(trainer_choice='Random',
                                                                    num_parties_jouees=10)
        self.assertTrue(is_consistent)

    def test_prepare_and_train_vs_monte_carlo(self):
        is_consistent_deterministic, _, _ = puissance4.prepare_and_train(trainer_choice='MC',
                                                                                  num_parties_jouees=10,
                                                                                  deterministic_root_action_sample=True)
        self.assertTrue(is_consistent_deterministic)

        is_consistent_random, _, _ = puissance4.prepare_and_train(trainer_choice='MC',
                                                                           num_parties_jouees=10,
                                                                           deterministic_root_action_sample=False)
        self.assertTrue(is_consistent_random)

    def test_prepare_and_train_vs_self(self):
        is_consistent, _, _ = puissance4.prepare_and_train(trainer_choice='UCT',
                                                                    num_parties_jouees=10)
        self.assertTrue(is_consistent)

    def test_save_then_load_model(self):
        is_consistent_save, _, _ = puissance4.prepare_and_train(trainer_choice='Random',
                                                                         num_parties_jouees=3,
                                                                         load_and_save_previously_trained_model=True)

        self.assertTrue(is_consistent_save)

        is_consistent_load, _, _ = puissance4.prepare_and_train(trainer_choice='MC',
                                                                         num_parties_jouees=3,
                                                                         load_and_save_previously_trained_model=True)

        self.assertTrue(is_consistent_load)


if __name__ == '__main__':
    unittest.main()



# Objective: check whether win conditions can be achieved in one step


def check_horizontale(grille, x, y):
    """Alignements horizontaux"""
    symbole = grille.grid[y][x]
    # Alignement horizontal de trois jetons consécutifs, le noeud (x,y) étant le plus à droite
    if grille.is_far_from_left(x):
        if all(symbole == grille.grid[y][x - i - 1] for i in range(2)):
            my_play = grille.play_if_possible(x - 3, y)
            if my_play is not None:
                return my_play
    # Alignements horizontaux, le noeud (x,y) étant le plus à gauche
    if grille.is_far_from_right(x):
        # Alignement horizontal de trois jetons consécutifs
        if all(symbole == grille.grid[y][x + i + 1] for i in range(2)):
            my_play = grille.play_if_possible(x + 3, y)
            if my_play is not None:
                return my_play
        # Alignement horizontal de trois jetons non consécutifs
        if symbole == grille.grid[y][x + 3]:
            # Alignement horizontal de la forme X.XX
            if symbole == grille.grid[y][x + 2]:
                my_play = grille.play_if_possible(x + 1, y)
                if my_play is not None:
                    return my_play
            # Alignement horizontal de la forme XX.X
            if symbole == grille.grid[y][x + 1]:
                my_play = grille.play_if_possible(x + 2, y)
                if my_play is not None:
                    return my_play

    return None


def check_verticale(grille, x, y):
    """Alignement vertical"""
    symbole = grille.grid[y][x]
    # Alignement vertical de trois jetons consécutifs, le noeud (x,y) étant le plus haut
    if grille.is_quite_far_from_bottom(y) and not (grille.is_at_top(y)):
        if all(symbole == grille.grid[y + i + 1][x] for i in range(2)):
            my_play = grille.play_if_possible(x, y - 1)
            if my_play is not None:
                return my_play
    return None


def check_oblique_montante(grille, x, y):
    """Alignements diagonaux montants (/) : allant du coin bas gauche au coin haut droit"""
    symbole = grille.grid[y][x]
    # Alignement diagonal montant de la forme XXX., le noeud (x,y) étant le plus bas et à gauche
    if grille.is_far_from_top(y) and grille.is_far_from_right(x):
        if all(symbole == grille.grid[y - i - 1][x + i + 1] for i in range(2)):
            my_play = grille.play_if_possible(x + 3, y - 2)
            if my_play is not None:
                return my_play
    # Alignements diagonaux montants, le noeud (x,y) étant le plus haut et à droite
    if grille.is_far_from_bottom(y) and grille.is_far_from_left(x):
        # Alignement diagonal de la forme .XXX
        if all(symbole == grille.grid[y + i + 1][x - i - 1] for i in range(2)):
            if grille.is_very_far_from_bottom(y):
                my_play = grille.play_if_possible(x - 3, y + 3)
                if my_play is not None:
                    return my_play
        if symbole == grille.grid[y + 3][x - 3]:
            # Alignement diagonal de la forme X.XX
            if symbole == grille.grid[y + 2][x - 2]:
                my_play = grille.play_if_possible(x - 1, y + 1)
                if my_play is not None:
                    return my_play
            # Alignement diagonal de la forme XX.X
            if symbole == grille.grid[y + 1][x - 1]:
                my_play = grille.play_if_possible(x - 2, y + 2)
                if my_play is not None:
                    return my_play

    return None


def check_oblique_descendante(grille, x, y):
    """Alignements diagonaux descendants (\) : allant du coin haut gauche au coin bas droit"""
    symbole = grille.grid[y][x]

    # Alignement diagonal descendant de la forme .XXX, le noeud (x,y) étant le plus bas et à droite
    if grille.is_far_from_top(y) and grille.is_far_from_left(x):
        if all(symbole == grille.grid[y - i - 1][x - i - 1] for i in range(2)):
            my_play = grille.play_if_possible(x - 2, y - 3)
            if my_play is not None:
                return my_play

    # Alignements diagonaux descendants, le noeud (x,y) étant le plus haut et à gauche

    if grille.is_far_from_bottom(y) and grille.is_far_from_right(x):
        # Alignement diagonal de la forme XXX.
        if all(symbole == grille.grid[y + i + 1][x + i + 1] for i in range(2)):
            if grille.is_very_far_from_bottom(y):
                my_play = grille.play_if_possible(x + 3, y + 3)
                if my_play is not None:
                    return my_play

        if symbole == grille.grid[y + 3][x + 3]:
            # Alignement diagonal de la forme X.XX
            if symbole == grille.grid[y + 2][x + 2]:
                my_play = grille.play_if_possible(x + 1, y + 1)
                if my_play is not None:
                    return my_play

            # Alignement diagonal de la forme XX.X
            if symbole == grille.grid[y + 1][x + 1]:
                my_play = grille.play_if_possible(x + 2, y + 2)
                if my_play is not None:
                    return my_play

    return None


def look_for_obvious_steps(grille, player_symbol='X', opponent_symbol='O'):
    """Vérifier s'il est possible de gagner pour l'un ou l'autre joueur.

    Rechercher un coup qui permette (d'abord) ou empêche (ensuite) un alignement de quatre jetons.
    Si oui, renvoyer le numéro de la colonne à jouer, sinon None"""

    # Check the player symbol first, so that a winning step is preferred to a step to avoid a loss
    for checked_symbol in [player_symbol, opponent_symbol]:
        for y in range(len(grille.grid)):
            for x in range(len(grille.grid[y])):
                if grille.grid[y][x] == checked_symbol:

                    my_play = check_horizontale(grille, x, y)
                    if my_play is not None:
                        return my_play

                    my_play = check_verticale(grille, x, y)
                    if my_play is not None:
                        return my_play

                    my_play = check_oblique_montante(grille, x, y)
                    if my_play is not None:
                        return my_play

                    my_play = check_oblique_descendante(grille, x, y)
                    if my_play is not None:
                        return my_play

    # Aucun coup urgent, alors renvoyer -1
    return None



from random import shuffle

from .agent.ai import AI
from .agent.mc import MC
from .agent.uct import UCT
from .env.grille import Grille
from .lib.utils import convert_to_column_display


def main():
    test_range_of_parameters()

    return True


def test_range_of_parameters(load_and_save_model=False):
    trainer_choice = 'UCT'  # One of the following texts: 'Random', 'MC', 'UCT'
    num_parties_jouees = 200

    results = dict()

    for num_tirages_MC in [8]:
        for num_descentes_dans_arbre in range(6, 14, 2):
            for facteur_uct in [0]:
                for max_num_steps_to_explore in [30]:
                    _, num_victories, num_steps = prepare_and_train(trainer_choice, num_parties_jouees,
                                                                    num_tirages_MC=num_tirages_MC,
                                                                    num_descentes_dans_arbre=num_descentes_dans_arbre,
                                                                    facteur_uct=facteur_uct,
                                                                    max_num_steps_to_explore=max_num_steps_to_explore,
                                                                    load_and_save_previously_trained_model=load_and_save_model)

                    num_victories['average_num_steps'] = sum(num_steps) / len(num_steps)
                    results[(num_tirages_MC, num_descentes_dans_arbre, facteur_uct, max_num_steps_to_explore)] \
                        = num_victories

                    print('Temporary summary: {}'.format(repr(results)))
                    print('\nAverage number of steps: {}'.format(num_victories['average_num_steps']))

    print('Final summary: {}'.format(repr(results)))

    return


# noinspection PyPep8Naming
def prepare_and_train(trainer_choice='MC', num_parties_jouees=3,
                      num_tirages_MC=None,
                      num_descentes_dans_arbre=None,
                      facteur_uct=None,
                      max_num_steps_to_explore=None,
                      bias_to_obvious_steps=None,
                      check_obvious_plays=None,
                      deterministic_root_action_sample=True,
                      enforce_identical_computing_resources=True,
                      load_and_save_previously_trained_model=False):
    # AI player which is learning by playing against the "trainer"
    learner = UCT()

    if num_tirages_MC is not None:
        learner.num_tirages_MC = num_tirages_MC

    if num_descentes_dans_arbre is not None:
        learner.num_descentes_dans_arbre = num_descentes_dans_arbre

    if facteur_uct is not None:
        learner.facteur_uct = facteur_uct

    if max_num_steps_to_explore is not None:
        learner.max_num_steps_to_explore = max_num_steps_to_explore

    if bias_to_obvious_steps is not None:
        learner.bias_to_obvious_steps = bias_to_obvious_steps

    if check_obvious_plays is not None:
        learner.check_obvious_plays = check_obvious_plays

    # AI player which is training the "learner"

    if trainer_choice == 'Random':
        trainer = AI()
    elif trainer_choice == 'MC':
        trainer = MC()
        trainer.deterministic_sampling_of_actions_at_root = deterministic_root_action_sample
    else:
        trainer = UCT()

    if enforce_identical_computing_resources:
        trainer.equalize_computing_resources(learner)

    # Load

    if load_and_save_previously_trained_model:
        learner.load_model()

    # Train
    learner, num_victories, num_victories_per_symbol, num_victories_per_player, num_steps = train(learner,
                                                                                                  trainer,
                                                                                                  num_parties_jouees,
                                                                                                  verbose=False)

    # Print

    learner.print()

    is_consistent = print_stats(num_victories, num_victories_per_symbol, num_victories_per_player)

    # Save

    if load_and_save_previously_trained_model:
        learner.save_model()

    return is_consistent, num_victories, num_steps


def train(learner, trainer, num_parties_jouees, verbose=False):
    num_victories_per_symbol = {'X': 0, 'O': 0, 'draw': 0}
    num_victories_per_player = {'learner': 0, 'trainer': 0, 'draw': 0}

    # X always starts.
    player_symbols = ['X', 'O']

    num_victories = {'draw': 0}
    for element in ['learner', 'trainer']:
        for symbol in player_symbols:
            num_victories[(element, symbol)] = 0

    # Decide which player is X and which is O.
    learner.player = player_symbols[0]
    trainer.player = player_symbols[1]

    learner.print()
    trainer.print()
    print()

    num_steps = []

    grille = Grille()

    for game_no in range(num_parties_jouees):

        grille.wipe()

        current_symbol = None
        current_player_name = None

        step_counter = 0
        while not (grille.check_victory()) and len(grille.look_for_allowed_steps()) > 0:

            if step_counter % 2 == 0:
                current_symbol = 'X'
            else:
                current_symbol = 'O'

            if current_symbol == learner.player:
                current_player = learner
                current_player_name = 'learner'
            else:
                current_player = trainer
                current_player_name = 'trainer'

            my_play, is_forced_play = current_player.play(grille)

            if verbose:
                if is_forced_play is not None and is_forced_play:
                    print(
                        'Obvious play by {} in ({}, {}) below.'.format(current_player_name, current_symbol,
                                                                       convert_to_column_display(my_play).upper()))

            grille.drop(current_player.get_player_symbol(), my_play)
            step_counter += 1

            if verbose:
                grille.show_grid()

        if grille.check_victory():
            winner_symbol = current_symbol
            winner_name = current_player_name
            num_victories[(winner_name, winner_symbol)] += 1
        else:
            winner_symbol = 'draw'
            winner_name = 'draw'
            num_victories['draw'] += 1

        num_victories_per_symbol[winner_symbol] += 1
        num_victories_per_player[winner_name] += 1

        print(
            'Game n°{}\tnum_steps = {}\twinner = {} ({})'.format(game_no + 1, step_counter, winner_name, winner_symbol))

        num_steps.append(step_counter)

        # Randomly shuffle symbols assigned to each player
        shuffle(player_symbols)

        # Decide which player is X and which is O.
        learner.player = player_symbols[0]
        trainer.player = player_symbols[1]

    return learner, num_victories, num_victories_per_symbol, num_victories_per_player, num_steps


def print_stats(num_victories, num_victories_per_symbol, num_victories_per_player):
    # Symbols: X vs. O

    num_games_for_symbols = num_victories_per_symbol['X'] \
                            + num_victories_per_symbol['O'] \
                            + num_victories_per_symbol['draw']

    try:
        ratio_victoires_per_symbol = (num_victories_per_symbol['X'] + 0.5 * num_victories_per_symbol['draw']) \
                                     / num_games_for_symbols
    except ZeroDivisionError:
        ratio_victoires_per_symbol = 0.5

    print("\n[Players X vs. O] winrate = {:.2f} (over {} games)".format(ratio_victoires_per_symbol,
                                                                        num_games_for_symbols))

    # Players: Learner vs. Trainer

    num_games_for_players = num_victories_per_player['learner'] \
                            + num_victories_per_player['trainer'] \
                            + num_victories_per_player['draw']

    try:
        ratio_victoires_per_player = (num_victories_per_player['learner'] + 0.5 * num_victories_per_player['draw']) \
                                     / num_games_for_players
    except ZeroDivisionError:
        ratio_victoires_per_player = 0.5

    print("\n[Learner vs. Trainer] winrate = {:.2f} (over {} games)".format(ratio_victoires_per_player,
                                                                            num_games_for_players))

    is_consistent = bool(num_games_for_symbols == num_games_for_players)

    # Additional print, useful to check everything works as intended, e.g. that the learnt model works for both X and O.
    print()
    print(num_victories)
    print(num_victories_per_symbol)
    print(num_victories_per_player)

    return is_consistent


if __name__ == "__main__":
    main()
